from geda.data_providers.base import DataProvider

from geda.utils.files import copy_files, save_yamls, create_dir, move
from pathlib import Path
from geda.utils.pylogger import get_pylogger
import scipy
from dataclasses import dataclass
from PIL import Image
from typing import Optional
import numpy as np
import cv2
import matplotlib.pyplot as plt
from tqdm.auto import tqdm

log = get_pylogger(__name__)

_URLS = {
    "images": "https://datasets.d2.mpi-inf.mpg.de/andriluka14cvpr/mpii_human_pose_v1.tar.gz",
    "annotations": "https://datasets.d2.mpi-inf.mpg.de/andriluka14cvpr/mpii_human_pose_v1_u12_2.zip",
}

LABELS = [
    "r_ankle",
    "r_knee",
    "r_hip",
    "l_hip",
    "l_knee",
    "l_ankle",
    "pelvis",
    "thorax",
    "upper_neck",
    "head_top",
    "r_wrist",
    "r_elbow",
    "r_shoulder",
    "l_shoulder",
    "l_elbosr",
    "l_wrist",
]

LIMBS = [
    [0, 1],  # R_ankle - R_knee
    [1, 2],  # R_knee - R_hip
    [2, 6],  # R_hip - pelvis
    [5, 4],  # L_ankle - L_knee
    [4, 3],  # L_knee - L_hip
    [3, 6],  # L_hip - pelvis
    [6, 7],  # pelvis - thorax
    [7, 8],  # thorax - upper_neck
    [8, 9],  # upper_neck - head_top
    [8, 12],  # upper_neck - R_shoulder
    [12, 11],  # R_shoulder - R_elbow
    [11, 10],  # R_elbow - R_wrist
    [8, 13],  # upper_neck - L_shoulder
    [13, 14],  # L_shoulder - L_elbow
    [14, 15],  # L_elbow - L_wrist
]


@dataclass
class PointAnnotation:
    x: int
    y: int
    id: int
    visibility: int

    @property
    def xy(self) -> tuple[int, int]:
        return self.x, self.y

    @classmethod
    def from_dict(cls, annot: dict) -> "PointAnnotation":
        return cls(**annot)


@dataclass
class PoseAnnotation:
    bbox: tuple[int, int, int, int]
    scale: float
    objpos_xy: tuple[int, int]
    head_xyxy: tuple[int, int, int, int]
    keypoints: list[PointAnnotation]

    @classmethod
    def from_dict(cls, annot: dict):
        bbox = annot["bbox"]
        scale = annot["scale"]
        objpos_xy = annot["objpos_xy"]
        head_xyxy = annot["head_xyxy"]
        keypoints = {
            point["id"]: PointAnnotation.from_dict(point)
            for point in annot["keypoints"]
        }
        return cls(bbox, scale, objpos_xy, head_xyxy, keypoints)


@dataclass
class Annotation:
    filename: str
    height: int
    width: int
    is_valid: bool
    poses: Optional[list[PoseAnnotation]] | None = None

    @classmethod
    def from_dict(cls, annot: dict):
        filename = annot["filename"]
        is_valid = annot["is_valid"]
        height = annot["height"]
        width = annot["width"]

        if not is_valid:
            return cls(filename, height, width, is_valid)
        poses = [PoseAnnotation.from_dict(obj_rect) for obj_rect in annot["objects"]]
        return cls(filename, height, width, is_valid, poses)

    def plot(self):
        if not self.is_valid or self.poses is None:
            return
        image = np.array(Image.open(f"images/{self.filename}"))
        for pose in self.poses:
            keypoints = pose.keypoints
            for _, kp in keypoints.items():
                color = (0, 128, 255) if kp.visibility == 2 else (128, 128, 128)
                cv2.circle(image, kp.xy, 3, color, -1)
            for id_1, id_2 in LIMBS:
                if id_1 not in keypoints or id_2 not in keypoints:
                    continue
                kp_1 = keypoints[id_1]
                kp_2 = keypoints[id_2]
                cv2.line(image, kp_1.xy, kp_2.xy, (50, 255, 50), 4)
        plt.imshow(image)


def parse_mpii_annotation(annot: np.ndarray, images_dir: str) -> dict:
    """
    returns annot dict in form
    visibility: 0 - not labeled, 1 - labeled but not visible, 2 - labeled and visible
    {
        'filename': string,
        'height': int, # TODO
        'width': int, # TODO
        'is_valid': bool, # False when there are no joint annotations
        'objects': [
            {
                'bbox': [int, int, int, int]
                'scale': float,
                'objpos_xy': [int, int],
                'head_xyxy': [int, int, int, int],
                'keypoints': [
                    {
                        'x': int,
                        'y': int,
                        'id': int,
                        'visibility': int
                    },
                    ...
                ]
            },
            ...
        ]
    }

    """
    filename = annot["image"]["name"].item()[0].item()
    annorects = annot["annorect"]
    if len(annorects) == 0:
        is_valid = False
    else:
        annorects = annorects[0]
        is_valid = (
            annorects.dtype.names is not None and "annopoints" in annorects.dtype.names
        )
    annot_dict = {"filename": filename, "is_valid": is_valid}
    if not is_valid:
        return annot_dict
    image = Image.open(f"{images_dir}/{filename}")
    annot_dict.update({"height": image.height, "width": image.width})

    person_dicts = []
    for ridx, annorect in enumerate(annorects):
        try:
            scale = annorect["scale"][0, 0].item()
        except IndexError:
            continue
        objpos_xy = [pos[0, 0].item() for pos in annorect["objpos"][0, 0]]
        head_xyxy = [annorect[coord][0, 0].item() for coord in ["x1", "y1", "x2", "y2"]]

        person_dict = {"scale": scale, "objpos_xy": objpos_xy, "head_xyxy": head_xyxy}

        annopoints = annorect["annopoints"]["point"][0, 0][0]

        # bbox is set as keypoints bounding rect
        joints_dicts = {}
        for point in annopoints:
            visibility = int(
                "is_visible" in point.dtype.names
                and [1] in point["is_visible"].tolist()
            )
            _id = point["id"][0, 0].item()
            joint_dict = {
                "x": int(point["x"][0, 0].item()),
                "y": int(point["y"][0, 0].item()),
                "id": _id,
                "visibility": visibility,
            }
            joints_dicts[_id] = joint_dict
        x_vals = [joint["x"] for joint in joints_dicts.values()]
        y_vals = [joint["y"] for joint in joints_dicts.values()]
        xmax = max(x_vals)
        ymax = max(y_vals)
        xmin = min(x_vals)
        ymin = min(y_vals)

        bbox_xc = int((xmax - xmin) // 2 + xmin)
        bbox_yc = int((ymax - ymin) // 2 + ymin)
        bbox_w = int(xmax - xmin)
        bbox_h = int(ymax - ymin)

        for _id in range(len(LABELS)):
            if _id not in joints_dicts:
                # filling it the same was as in COCO
                joints_dicts[_id] = {"x": 0, "y": 0, "id": _id, "visibility": 0}

        person_dict["bbox"] = [bbox_xc, bbox_yc, bbox_w, bbox_h]
        keypoints = sorted(list(joints_dicts.values()), key=lambda kp: kp["id"])
        person_dict["keypoints"] = keypoints
        person_dicts.append(person_dict)
    annot_dict["objects"] = person_dicts
    return annot_dict


class MPIIDataProvider(DataProvider):
    URL = "http://human-pose.mpi-inf.mpg.de/"

    def __init__(self, root: str):
        self.task = "HumanPose"
        self.task_root = f"{root}/{self.task}"
        super().__init__(urls=_URLS, root=root)

    def move_to_raw_root(self):
        src_annots_path = f"{self.root}/mpii_human_pose_v1_u12_2"
        src_data_path = f"{self.root}/images"
        move(src_annots_path, f"{self.raw_root}/annots")
        move(src_data_path, f"{self.raw_root}/images")

    def _get_filepaths(self, dirnames: list[str] = ["annots", "images"]):
        super()._get_filepaths(dirnames)

    def set_annotations(self):
        annots_struct = scipy.io.loadmat(
            f"{self.raw_root}/annots/mpii_human_pose_v1_u12_1.mat"
        )
        release = annots_struct["RELEASE"][0, 0]
        annolist = release["annolist"][0]
        # train/val/test splits are determined on annotations validity, not on "img_train" field
        is_train = release["img_train"][0]  # not used
        annots = [parse_mpii_annotation(a, f"{self.raw_root}/images") for a in annolist]
        self.annots = annots

    def _get_split_ids(self):
        self.set_annotations()
        valid = [a["is_valid"] for a in self.annots]

        train_val_idxs = np.where(valid)[0]
        test_idxs = np.where(~np.array(valid))[0]

        train_ratio = 0.8
        train_n_samples = int(train_ratio * len(train_val_idxs))
        train_idxs = train_val_idxs[:train_n_samples]
        val_idxs = train_val_idxs[train_n_samples:]

        return {
            "train": train_idxs.tolist(),
            "val": val_idxs.tolist(),
            "test": test_idxs.tolist(),
        }

    def arrange_files(self):
        Path(self.task_root).mkdir(parents=True, exist_ok=True)

        for split, idxs in self.split_ids.items():
            split_annots = [self.annots[idx] for idx in idxs]
            names = ["annots", "images"]
            dst_paths = {
                name: create_dir(Path(self.task_root) / name / split) for name in names
            }
            src_imgs_path = f"{self.raw_root}/images"
            dst_imgs_path = dst_paths["images"]
            dst_annots_path = dst_paths["annots"]

            filenames = [annot["filename"] for annot in split_annots]
            src_imgs_filepaths = [f"{src_imgs_path}/{fname}" for fname in filenames]
            dst_imgs_filepaths = [f"{dst_imgs_path}/{fname}" for fname in filenames]

            log.info(f"Copying {split} images from {src_imgs_path} to {dst_imgs_path}")
            copy_files(src_imgs_filepaths, dst_imgs_filepaths)

            log.info(f"Saving {split} annotations as .yaml files in {dst_annots_path}")
            yaml_paths = [
                f'{dst_annots_path}/{annot["filename"].replace(".jpg", "")}.yaml'
                for annot in split_annots.values()
            ]
            save_yamls(split_annots, yaml_paths)


if __name__ == "__main__":
    from geda.utils.config import ROOT

    root = str(ROOT / "data" / "MPII")
    dp = MPIIDataProvider(root)
    dp.get_data()



from tqdm import tqdm
import urllib.request
import tarfile
from geda.utils.pylogger import get_pylogger
import os
import shutil
from pathlib import Path
import zipfile
import yaml
import gzip
from joblib import Parallel, delayed

log = get_pylogger(__name__)

_filepaths = list[str]


class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        return self.update(b * bsize - self.n)  # also sets self.n = b * bsize


def download_file(url: str, filepath: str):
    log.info(f"Downloading {url} to {filepath}")
    with TqdmUpTo(
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
        miniters=1,
        desc=url.split("/")[-1],
    ) as t:  # all optional kwargs
        urllib.request.urlretrieve(
            url, filename=filepath, reporthook=t.update_to, data=None
        )
        t.total = t.n
    log.info("Download finished")


def unzip_tar(filepath: str, dst_path, mode: str = "r"):
    with tarfile.open(filepath, mode) as tar:
        tar.extractall(dst_path)
    log.info("Unzipping finished")


def unzip_zip(filepath: str, dst_path: str):
    with zipfile.ZipFile(filepath, "r") as zip_ref:
        zip_ref.extractall(dst_path)


def unzip_gz(filepath: str):
    with gzip.open(filepath, "rb") as f_in:
        with open(filepath.replace(".gz", ""), "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)


def unzip(filepath: str, dst_path: str, remove: bool = False):
    log.info(f"Unzipping {filepath} to {dst_path}.")
    ext = filepath.split(".")[-1]
    if ext == "tar":
        unzip_tar(filepath, dst_path)
    elif ext == "zip":
        unzip_zip(filepath, dst_path)
    elif ext == "gz":
        unzip_gz(filepath)
    if remove:
        os.remove(filepath)
        log.info(f"Removed {filepath}")


def unzip_many(filepaths: list[str], dst_paths: list[str] | str, remove: bool = False):
    desc = "Unzipping files"
    total = len(filepaths)
    if isinstance(dst_paths, str):
        dst_paths = [dst_paths] * total
    Parallel(n_jobs=-1)(
        delayed(unzip)(zip_filepath, dst_path, remove)
        for zip_filepath, dst_path in tqdm(
            zip(filepaths, dst_paths), desc=desc, total=total
        )
    )


def save_txt_to_file(txt: str, filename: str):
    with open(filename, "w") as file:
        file.write(txt)


def read_txt_file(filepath: str) -> list[str]:
    with open(filepath, "r") as file:
        lines = file.readlines()
        lines = [
            line.strip() for line in lines
        ]  # Optional: Remove leading/trailing whitespace
    return lines


def add_prefix_to_files(directory: str, prefix: str, ext: str = ".png"):
    log.info(f"Adding {prefix} prefix to all {ext} files in {directory} directory")

    for filename in os.listdir(directory):
        if os.path.isfile(os.path.join(directory, filename)) and filename.endswith(ext):
            new_filename = prefix + filename
            os.rename(
                os.path.join(directory, filename), os.path.join(directory, new_filename)
            )
    log.info("Prefix addition finished")


def move(src: str, dst: str):
    shutil.move(src, dst)
    log.info(f"Moved {src} to {dst}")


def move_many(sources: _filepaths, destinations: _filepaths):
    total = len(sources)
    desc = "Moving files"
    log.info(f"Moving {total} files/directories")
    Parallel(n_jobs=-1)(
        delayed(shutil.move)(src, dst)
        for src, dst in tqdm(zip(sources, destinations), desc=desc, total=total)
    )
    log.info(f"Moved {total} files")


def copy_directory(src_dir: str, dst_dir: str):
    shutil.copytree(src_dir, dst_dir)
    log.info(f"Copied {src_dir} to {dst_dir}")


def remove_directory(dirpath: str):
    shutil.rmtree(dirpath)
    log.info(f"Removed {dirpath} directory")


def path_exists(path: str):
    return os.path.exists(path)


def copy_file(src: str, dst: str):
    try:
        shutil.copy2(src, dst)
    except FileNotFoundError as e:
        log.warn(f"{src} not found")


def copy_all_files(src_dir: Path, dst_dir: Path, ext: str = ".png"):
    def _copy(filename: str):
        if filename.lower().endswith(ext):
            source = str(src_dir / filename)
            destination = str(dst_dir / filename)
            copy_file(source, destination)

    desc = "Copying files"
    filenames = os.listdir(src_dir)
    Parallel(n_jobs=-1)(delayed(_copy)(fname) for fname in tqdm(filenames, desc=desc))
    log.info(f"Copied all {ext} files ({len(filenames)}) from {src_dir} to {dst_dir}")


def copy_files(src_filepaths: _filepaths, dst_filepaths: _filepaths):
    desc = "Copying files"
    total = len(src_filepaths)
    Parallel(n_jobs=-1)(
        delayed(copy_file)(src, dst)
        for src, dst in tqdm(zip(src_filepaths, dst_filepaths), desc=desc, total=total)
    )
    log.info(f"Copied {total} files")


def create_dir(path: Path, return_str: bool = True):
    path.mkdir(parents=True, exist_ok=True)
    if return_str:
        return str(path)
    return path


def load_yaml(path: Path | str) -> dict:
    with open(path, "r") as file:
        yaml_dct = yaml.safe_load(file)
    return yaml_dct


def save_yaml(dct: dict, path: Path | str):
    with open(path, "w") as file:
        yaml.dump(dct, file)


def save_yamls(dcts: list[dict], paths: list[str]):
    desc = "Copying yaml files"
    total = len(dcts)
    Parallel(n_jobs=-1)(
        delayed(save_yaml)(dct, path)
        for dct, path in tqdm(zip(dcts, paths), desc=desc, total=total)
    )



from geda.utils.files import (
    download_file,
    unzip_many,
    path_exists,
    load_yaml,
    save_yaml,
)
from geda.utils.pylogger import get_pylogger
from geda.utils.images import load_img_to_array
from abc import abstractmethod
from pathlib import Path
import glob
import numpy as np
from tqdm.auto import tqdm
from functools import cached_property
from typing import Literal

log = get_pylogger(__name__)


class DataProvider:
    URL: str
    zip_filepaths: list[str]

    def __init__(self, urls: dict[str, str], root: str):
        self.urls = urls
        self.root = root
        if not hasattr(self, "task_root"):
            self.task_root = "Task"
        Path(self.root).mkdir(parents=True, exist_ok=True)
        self.raw_root = f"{root}/raw"

    def get_data(self, remove_zip: bool = False):
        if not self.check_if_present():
            self.zip_filepaths = self.download()  # TODO
            self.unzip(remove=remove_zip)
            log.info("Loading splits ids")
            self.split_ids = self._get_split_ids()
            log.info("Arranging files")
            self.arrange_files()
            self.create_labels()
            self.filepaths = self._get_filepaths()
        else:
            log.info("Canceling data download and files rearranging")
            self.split_ids = self._get_split_ids()
            self.filepaths = self._get_filepaths()

    def download(self):
        if path_exists(self.raw_root):
            log.info(f"{self.raw_root} is already present. Download canceled")
            return
        zip_filepaths = []
        for name, url in self.urls.items():
            ext = url.split("/")[-1].split(".")[1]
            zip_filepath = f"{self.root}/{name}.{ext}"
            if ext in ["zip", "tar", "gz"]:
                zip_filepaths.append(zip_filepath)
            if path_exists(zip_filepath):
                log.info(f"{zip_filepath} is already present. Download canceled")
            else:
                download_file(url, zip_filepath)
        return zip_filepaths

    def unzip(self, remove: bool = False):
        if path_exists(self.raw_root):
            log.info(f"{self.raw_root} is already present. Unzip canceled")
            return
        Path(self.raw_root).mkdir(parents=True, exist_ok=True)
        unzip_many(self.zip_filepaths, self.root, remove)
        self.move_to_raw_root()
        if remove:
            self.zip_filepaths.clear()

    def check_if_present(self) -> bool:
        is_present = path_exists(self.task_root)
        if is_present:
            log.info(f"{self.task_root} is already present")
        return is_present

    @abstractmethod
    def move_to_raw_root(self):
        raise NotImplementedError()

    @abstractmethod
    def _get_split_ids(self):
        raise NotImplementedError()

    @property
    def splits(self) -> list[str]:
        return list(self.split_ids.keys())

    @abstractmethod
    def arrange_files(self):
        raise NotImplementedError()

    def _get_filepaths(self, dirnames: list[str] = []):
        filepaths = {}
        for dirname in dirnames:
            splits_paths = {}
            for split in self.split_ids:
                paths = glob.glob(f"{self.task_root}/{dirname}/{split}/*")
                splits_paths[split] = sorted(paths)
            filepaths[dirname] = splits_paths
        return filepaths

    @abstractmethod
    def create_labels(self):
        log.warn(f"create_labels method not implemented for {self.__class__.__name__}")


class SegmentationDataProvider(DataProvider):
    def __init__(
        self,
        urls: dict[str, str],
        root: str,
        labels_format: Literal["yolo"] | None = "yolo",
    ):
        super().__init__(urls, root)
        self.labels_format = labels_format

    def _get_filepaths(self, dirnames: list[str] = ["masks", "images", "labels"]):
        super()._get_filepaths(dirnames)

    def _get_class_counts(
        self, masks_dirname: str = "masks"
    ) -> dict[str, dict[str, int]]:
        class_counts_path = f"{self.task_root}/class_counts.yaml"
        if path_exists(class_counts_path):
            log.info(f"{class_counts_path} is present. Loading counts from that file.")
            return load_yaml(class_counts_path)
        splits_cls_freqs = {}
        for split in self.splits:
            masks_filepaths = self.filepaths[masks_dirname][split]
            classes_counts = {}
            for path in tqdm(
                masks_filepaths, desc=f"Counting class pixels for {split} split"
            ):
                mask = load_img_to_array(path)
                classes, counts = np.unique(mask, return_counts=True)
                for _class, count in zip(classes, counts):
                    if _class not in classes_counts:
                        classes_counts[_class.item()] = count.item()
                    else:
                        classes_counts[_class.item()] += count.item()
            splits_cls_freqs[split] = classes_counts
        save_yaml(splits_cls_freqs, class_counts_path)
        return splits_cls_freqs

    def _get_class_frequencies(self) -> dict[str, dict[str, float]]:
        class_frequencies_path = f"{self.task_root}/class_frequencies.yaml"
        if path_exists(class_frequencies_path):
            log.info(
                f"{class_frequencies_path} is present. Loading frequencies from that file."
            )
            return load_yaml(class_frequencies_path)
        splits_cls_freqs = {}
        for split in self.splits:
            split_cls_counts = self.class_counts[split]
            n_total_pixels = sum(split_cls_counts.values())
            cls_freqs = {k: v / n_total_pixels for k, v in split_cls_counts.items()}
            splits_cls_freqs[split] = cls_freqs
        save_yaml(splits_cls_freqs, class_frequencies_path)
        return splits_cls_freqs

    def _set_id2class(self, id2class: dict[int, str]):
        self._id2class = id2class

    @cached_property
    def class_counts(self):
        return self._get_class_counts()

    @cached_property
    def id2class(self):
        return self._id2class

    @cached_property
    def class_frequencies(self):
        return self._get_class_frequencies()

    @cached_property
    def class2id(self):
        return {v: k for k, v in self.id2class.items()}

    def save_id2class(self):
        save_yaml(self.id2class, f"{self.task_root}/id2class.yaml")

    def get_data(self):
        super().get_data()
        self.save_id2class()
        self._get_class_counts()
        self._get_class_frequencies()


class ClassificationDataProvider(DataProvider):
    def _get_filepaths(self, dirnames: list[str] = ["images", "labels"]):
        return super()._get_filepaths(dirnames)



from geda.data_providers.base import DataProvider

from geda.utils.files import copy_files, save_yamls, create_dir, move_many

from pathlib import Path
from geda.utils.pylogger import get_pylogger
from dataclasses import dataclass
from PIL import Image
from typing import Optional
import numpy as np
import cv2
import matplotlib.pyplot as plt
import json
import glob
from tqdm.auto import tqdm

log = get_pylogger(__name__)

_URLS = {
    "train2017": "http://images.cocodataset.org/zips/train2017.zip",
    "val2017": "http://images.cocodataset.org/zips/val2017.zip",
    "test2017": "http://images.cocodataset.org/zips/test2017.zip",
    "annotations_trainval2017": "http://images.cocodataset.org/annotations/annotations_trainval2017.zip",
}

LABELS = [
    "nose",
    "left_eye",
    "right_eye",
    "left_ear",
    "right_ear",
    "left_shoulder",
    "right_shoulder",
    "left_elbow",
    "right_elbow",
    "left_wrist",
    "right_wrist",
    "left_hip",
    "right_hip",
    "left_knee",
    "right_knee",
    "left_ankle",
    "right_ankle",
]

LIMBS = [
    (9, 7),
    (7, 5),
    (5, 3),
    (3, 1),
    (1, 0),
    (0, 2),
    (1, 2),
    (2, 4),
    (4, 6),
    (6, 8),
    (8, 10),
    (5, 6),
    (5, 11),
    (6, 12),
    (11, 12),
    (11, 13),
    (13, 15),
    (12, 14),
    (14, 16),
]


@dataclass
class PointAnnotation:
    x: int
    y: int
    id: int
    visibility: bool

    @property
    def xy(self) -> tuple[int, int]:
        return self.x, self.y

    @classmethod
    def from_dict(cls, annot: dict) -> "PointAnnotation":
        return cls(**annot)


@dataclass
class PoseAnnotation:
    bbox: tuple[int, int, int, int]
    keypoints: dict[int, PointAnnotation]

    @classmethod
    def from_dict(cls, annot: dict):
        bbox = annot["bbox"]
        keypoints = {
            point["id"]: PointAnnotation.from_dict(point)
            for point in annot["keypoints"]
        }
        return cls(bbox, keypoints)


@dataclass
class Annotation:
    filename: str
    height: int
    width: int
    poses: Optional[list[PoseAnnotation]] | None = None

    @classmethod
    def from_dict(cls, annot: dict):
        filename = annot["filename"]
        height = annot["height"]
        width = annot["width"]
        poses = [PoseAnnotation.from_dict(obj_rect) for obj_rect in annot["objects"]]
        return cls(filename, height, width, poses)

    def plot(self):
        if self.poses is None:
            return
        image = np.array(Image.open(f"images/{self.filename}"))
        for pose in self.poses:
            keypoints = pose.keypoints
            for _, kp in keypoints.items():
                color = (0, 128, 255) if kp.visibility == 2 else (128, 128, 128)
                cv2.circle(image, kp.xy, 3, color, -1)
            for id_1, id_2 in LIMBS:
                if id_1 not in keypoints or id_2 not in keypoints:
                    continue
                kp_1 = keypoints[id_1]
                kp_2 = keypoints[id_2]

                cv2.line(image, kp_1.xy, kp_2.xy, (50, 255, 50), 4)
        image = cv2.resize(image, (0, 0), fx=2, fy=2)
        plt.figure(figsize=(12, 24))
        plt.imshow(image)


def parse_coco_keypoint_annot(annots: list[dict], info: dict) -> dict:
    annot_dict = info
    person_dicts = []
    for annot in annots:
        person_dict = {
            "bbox": annot["bbox"],
            "iscrowd": annot["iscrowd"],
            "num_keypoints": annot["num_keypoints"],
            "segmentation": annot["segmentation"],
        }
        kpts = annot["keypoints"]
        joint_dicts = []
        for i, (x, y, v) in enumerate(zip(kpts[::3], kpts[1::3], kpts[2::3])):
            joint_dict = {"x": int(x), "y": int(y), "id": i, "visibility": int(v)}
            joint_dicts.append(joint_dict)
        person_dict["keypoints"] = joint_dicts
        person_dicts.append(person_dict)
    annot_dict["objects"] = person_dicts
    return annot_dict


class COCOKeypointsDataProvider(DataProvider):
    URL = "https://cocodataset.org/#download"

    def __init__(self, root: str):
        self.task = "HumanPose"
        self.task_root = f"{root}/{self.task}"
        super().__init__(urls=_URLS, root=root)

    def move_to_raw_root(self):
        dirnames = ["train2017", "val2017", "test2017", "annotations"]
        src_paths = [f"{self.root}/{dirname}" for dirname in dirnames]
        dst_paths = [f"{self.raw_root}/{dirname}" for dirname in dirnames]
        move_many(src_paths, dst_paths)

    def _get_filepaths(self, dirnames: list[str] = ["annots", "images"]):
        super()._get_filepaths(dirnames)

    def set_annotations(self):
        annots_dir = f"{self.raw_root}/annotations"
        annot_splits = {"train": {}, "val": {}}
        for split in annot_splits:
            with open(f"{annots_dir}/person_keypoints_{split}2017.json") as f:
                annots = json.load(f)

            id2info = {}
            for info in annots["images"]:
                id2info[info["id"]] = {
                    "filename": info["file_name"],
                    "height": info["height"],
                    "width": info["width"],
                }

            id2annots = {}

            for annot in annots["annotations"]:
                img_id = annot["image_id"]
                if img_id not in id2annots:
                    id2annots[img_id] = [annot]
                else:
                    id2annots[img_id].append(annot)

            for _id, annots in id2annots.items():
                id2annots[_id] = parse_coco_keypoint_annot(annots, id2info[_id])
            annot_splits[split] = id2annots
        test_filepaths = sorted(glob.glob(f"{self.raw_root}/test2017/*"))
        test_id2annots = {}
        for filepath in test_filepaths:
            filename = filepath.split("/")[-1]
            _id = filename.split(".")[0]
            image = Image.open(filepath)
            test_id2annots[_id] = {
                "filename": filename,
                "height": image.height,
                "width": image.width,
            }
        annot_splits["test"] = test_id2annots
        self.annots = annot_splits

    def _get_split_ids(self):
        self.set_annotations()

        return {
            split: list(split_annots.keys())
            for split, split_annots in self.annots.items()
        }

    def arrange_files(self):
        Path(self.task_root).mkdir(parents=True, exist_ok=True)
        for split, annots in tqdm(self.annots.items(), desc="Splits"):
            names = ["annots", "images"]
            dst_paths = {
                name: create_dir(Path(self.task_root) / name / split) for name in names
            }
            src_imgs_path = f"{self.raw_root}/{split}2017"
            dst_imgs_path = dst_paths["images"]

            dst_annots_path = dst_paths["annots"]

            filenames = [annot["filename"] for _id, annot in annots.items()]
            src_imgs_filepaths = [f"{src_imgs_path}/{fname}" for fname in filenames]
            dst_imgs_filepaths = [f"{dst_imgs_path}/{fname}" for fname in filenames]

            log.info(f"Copying {split} images from {src_imgs_path} to {dst_imgs_path}")
            copy_files(src_imgs_filepaths, dst_imgs_filepaths)

            log.info(f"Saving {split} annotations as .yaml files in {dst_annots_path}")
            yaml_paths = [
                f'{dst_annots_path}/{annot["filename"].replace(".jpg", "")}.yaml'
                for annot in annots.values()
            ]
            save_yamls(list(annots.values()), yaml_paths)


if __name__ == "__main__":
    from geda.utils.config import ROOT

    root = str(ROOT / "data" / "COCO")
    dp = COCOKeypointsDataProvider(root)
    dp.get_data()



from typing import Literal
import geda.data_providers as gdp

DATA_PROVIDERS = {
    "MNIST": gdp.mnist.MNISTDataProvider,
    "DUTS": gdp.duts.DUTSDataProvider,
    "NYUDv2": gdp.nyud.NYUDv2DataProvider,
    "VOC_InstanceSegmentation": gdp.voc.VOCInstanceSegmentationDataProvider,
    "VOC_SemanticSegmentation": gdp.voc.VOCSemanticSegmentationDataProvider,
    "VOC_PersonPartSegmentation": gdp.voc.VOCPersonPartSegmentationDataProvider,
    "VOC_SemanticSegmentationAug": gdp.voc.VOCSemanticSegmentationAugDataProvider,
    "VOC_Main": gdp.voc.VOCMainDataProvider,
    "VOC_Action": gdp.voc.VOCActionDataProvider,
    "VOC_Layout": gdp.voc.VOCLayoutDataProvider,
    "MPII": gdp.mpii.MPIIDataProvider,
    "COCO_Keypoints": gdp.coco.COCOKeypointsDataProvider,
}

DATA_PROVIDERS_NAMES = Literal[
    "MNIST",
    "DUTS",
    "NYUDv2",
    "VOC_InstanceSegmentation",
    "VOC_SemanticSegmentation",
    "VOC_SemanticSegmentationAug",
    "VOC_PersonPartSegmentation",
    "VOC_Main",
    "VOC_Action",
    "VOC_Layout",
    "MPII",
    "COCO_Keypoints",
]


def get_data(name: DATA_PROVIDERS_NAMES, root: str, **kwargs) -> gdp.base.DataProvider:
    data_provider: gdp.base.DataProvider = DATA_PROVIDERS[name](root=root)
    data_provider.get_data()
    return data_provider



from geda.data_providers.coco import COCOKeypointsDataProvider
from geda.datasets.base import BaseKeypointsDataset
from typing import Literal


class COCOKeypointsDataset(BaseKeypointsDataset):
    def __init__(self, root: str, split: Literal["train", "val", "test"]):
        data_provider = COCOKeypointsDataProvider(root)
        super().__init__(root, split, data_provider)


if __name__ == "__main__":
    from geda.utils.config import ROOT

    root = str(ROOT / "data" / "DUTS")
    train_ds = COCOKeypointsDataset(root, "train")
    val_ds = COCOKeypointsDataset(root, "val")
    test_ds = COCOKeypointsDataset(root, "test")



from geda.utils.colors import color_map
import png
import numpy as np
import math
from PIL import Image


def load_img_to_array(filepath: str) -> np.ndarray:
    if filepath.endswith((".jpg", "png", "jpeg")):
        return np.array(Image.open(filepath))
    elif filepath.endswith(".npy"):
        return np.load(filepath)
    else:
        raise Exception("Wrong file extension. .jpg, .png, .jpeg and .npy are supported.")


def save_gray_array_as_color_png(
    array: np.ndarray, palette: list[tuple[int, int, int]] | None, filename: str = "gray2color.png"
):
    height, width = array.shape
    bitdepth = math.ceil(math.log2(len(palette)))
    if palette is None:
        palette = color_map(256)
    w = png.Writer(width, height, palette=palette, bitdepth=bitdepth)
    f = open(filename, "wb")
    w.write(f, array.tolist())



from geda.data_providers.base import SegmentationDataProvider

from geda.utils.files import (
    move,
    remove_directory,
    read_txt_file,
    copy_files,
    create_dir,
    unzip,
    path_exists,
    download_file,
)
from geda.utils.colors import color_map
from geda.parsers.yolo import parse_segmentation_masks_to_yolo
from typing import Literal
from pathlib import Path
import glob
from geda.utils.pylogger import get_pylogger


log = get_pylogger(__name__)

_TRAINAUG_IDS_URL = "https://gist.githubusercontent.com/sun11/2dbda6b31acc7c6292d14a872d0c90b7/raw/5f5a5270089239ef2f6b65b1cc55208355b5acca/trainaug.txt"

_SBD_URL = {"trainaug": "https://www.dropbox.com/s/oeu149j8qtbs1x0/SegmentationClassAug.zip?dl=0"}

_URLS = {
    2007: {
        "trainval_2007": "http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtrainval_06-Nov-2007.tar",
        "test_2007": "http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtest_06-Nov-2007.tar",
    },
    2012: {
        "trainval_2012": "http://host.robots.ox.ac.uk/pascal/VOC/voc2012/VOCtrainval_11-May-2012.tar",
    },
}

_PERSON_PART_URL = {"person_part": "http://www.liangchiehchen.com/data/pascal_person_part.zip"}

_SEG_SEMANTIC_LABELS = [
    "aeroplane",
    "bicycle",
    "bird",
    "boat",
    "bottle",
    "bus",
    "car",
    "cat",
    "chair",
    "cow",
    "diningtable",
    "dog",
    "horse",
    "motorbike",
    "person",
    "pottedplant",
    "sheep",
    "sofa",
    "train",
    "tvmonitor",
]
_SEG_PERSON_PART_LABELS = ["Head", "Torso", "Upper Arms", "Lower Arms", "Upper Legs", "Lower Legs"]
_SEG_INSTANCE_LABELS = [str(i) for i in range(1, 256)]

_SEG_SEMANTIC_ID2LABEL = {i: label for i, label in enumerate(_SEG_SEMANTIC_LABELS, start=1)} | {
    0: "background",
    255: "void",
}

_SEG_PERSON_PART_ID2LABEL = {
    i: label for i, label in enumerate(_SEG_PERSON_PART_LABELS, start=1)
} | {0: "background", None: "void"}


_SEG_INSTANCE_ID2LABEL = {i: label for i, label in enumerate(_SEG_INSTANCE_LABELS, start=1)} | {
    0: "background",
    255: "void",
}

_SEG_ID2LABEL = {
    "semantic": _SEG_SEMANTIC_ID2LABEL,
    "semantic_aug": _SEG_SEMANTIC_ID2LABEL,
    "person_part": _SEG_PERSON_PART_ID2LABEL,
    "instance": _SEG_INSTANCE_ID2LABEL,
}

_PALETTE = color_map(256)


class VOCDataProvider(SegmentationDataProvider):
    URL = "http://host.robots.ox.ac.uk/pascal/VOC/"

    def __init__(
        self,
        root: str,
        task: Literal[
            "Action",
            "Layout",
            "Main",
            "SegmentationClass",
            "SegmentationClassAug",
            "SegmentationObject",
            "SegmentationPersonPart",
        ],
        labels_format: Literal["yolo"] | None = "yolo",
    ):
        self.year = 2012
        self.task_root = f"{root}/{task}"
        if task in ["SegmentationClass", "SegmentationObject"]:
            self.task = "Segmentation"
        else:
            self.task = task
        urls = _URLS[2012]
        if task == "SegmentationPersonPart":
            urls.update(_PERSON_PART_URL)
        # elif task == "SegmentationClassAug": # TODO: dropbox requires authentication
        #     urls.update(_SBD_URL)
        super().__init__(urls=urls, root=root, labels_format=labels_format)

    def move_to_raw_root(self):
        src_dir = f"{self.root}/VOCdevkit/VOC{self.year}"
        for directory in glob.glob(f"{src_dir}/*"):
            move(directory, self.raw_root)
        remove_directory(f"{self.root}/VOCdevkit")

    def _get_split_ids(self):
        task_dir = f"{self.raw_root}/ImageSets/{self.task}"
        id_file_suffix = "_id" if self.task == "SegmentationPersonPart" else ""
        train_ids = read_txt_file(f"{task_dir}/train{id_file_suffix}.txt")
        val_ids = read_txt_file(f"{task_dir}/val{id_file_suffix}.txt")
        if self.task == "Layout":
            train_ids = [_id.split(" ")[0] for _id in train_ids]
            val_ids = [_id.split(" ")[0] for _id in val_ids]
            train_ids = sorted(list(set(train_ids)))
            val_ids = sorted(list(set(val_ids)))
        split_ids = {"train": train_ids, "val": val_ids}
        if self.task == "SegmentationClassAug":
            split_ids["trainaug"] = read_txt_file(f"{task_dir}/trainaug.txt")
        return split_ids

    def arrange_files(self):
        for split, ids in self.split_ids.items():
            names = ["annots", "images", "labels"]
            dst_paths = {name: create_dir(Path(self.task_root) / name / split) for name in names}
            src_paths = {
                "annots": f"{self.raw_root}/Annotations",
                "images": f"{self.raw_root}/JPEGImages",
            }

            src_annots_filepaths = [f"{src_paths['annots']}/{_id}.xml" for _id in ids]
            src_images_filepaths = [f"{src_paths['images']}/{_id}.jpg" for _id in ids]

            dst_annots_filepaths = [f"{dst_paths['annots']}/{_id}.xml" for _id in ids]
            dst_images_filepaths = [f"{dst_paths['images']}/{_id}.jpg" for _id in ids]

            log.info(f"Moving {split} annots from {src_paths['annots']} to {dst_paths['annots']}")
            log.info(f"Moving {split} images from {src_paths['images']} to {dst_paths['images']}")
            copy_files(src_annots_filepaths, dst_annots_filepaths)
            copy_files(src_images_filepaths, dst_images_filepaths)


class VOCMainDataProvider(VOCDataProvider):
    def __init__(self, root: str, labels_format: Literal["yolo"] = "yolo"):
        super().__init__(root=root, task="Main", labels_format=labels_format)


class VOCActionDataProvider(VOCDataProvider):
    def __init__(self, root: str, labels_format: Literal["yolo"] = "yolo"):
        super().__init__(root=root, task="Action", labels_format=labels_format)


class VOCLayoutDataProvider(VOCDataProvider):
    def __init__(self, root: str, labels_format: Literal["yolo"] = "yolo"):
        super().__init__(root=root, task="Layout", labels_format=labels_format)


class VOCSegmentationDataProvider(VOCDataProvider):
    def __init__(
        self,
        root: str,
        mode: Literal["semantic", "semantic_aug", "instance", "person_part"] = "semantic",
        labels_format: Literal["yolo"] | None = "yolo",
    ):
        self.mode = mode
        self._set_id2class(id2class=_SEG_ID2LABEL[mode])
        if mode == "semantic":
            mask_dirname = "SegmentationClass"
        elif mode == "semantic_aug":
            mask_dirname = "SegmentationClassAug"
        elif mode == "instance":
            mask_dirname = "SegmentationObject"
        elif mode == "person_part":
            mask_dirname = "SegmentationPersonPart"
        self.mask_dirname = mask_dirname
        super().__init__(root=root, task=mask_dirname, labels_format=labels_format)

    def arrange_files(self):
        super().arrange_files()
        for split, ids in self.split_ids.items():
            src_masks_path = f"{self.raw_root}/{self.mask_dirname}"
            dst_masks_path = create_dir(Path(self.task_root) / "masks" / split)

            src_filepaths = [f"{src_masks_path}/{_id}.png" for _id in ids]
            dst_filepaths = [f"{dst_masks_path}/{_id}.png" for _id in ids]

            log.info(f"Moving {split} masks from {src_masks_path} to {dst_masks_path}")
            copy_files(src_filepaths, dst_filepaths)

    def _get_filepaths(self):
        return super()._get_filepaths(["annots", "images", "labels", "masks"])

    def create_labels(self):
        masks_filepaths = sorted(glob.glob(f"{self.task_root}/masks/*/*"))
        if self.labels_format == "yolo":
            parse_segmentation_masks_to_yolo(
                masks_filepaths,
                background=self.class2id["background"],
                contour=self.class2id["void"],
            )
        else:
            log.warn(f"Only YOLO label format is implemented ({self.labels_format} passed)")


class VOCInstanceSegmentationDataProvider(VOCSegmentationDataProvider):
    def __init__(self, root: str, labels_format: Literal["yolo"] | None = None):
        super().__init__(root=root, mode="instance", labels_format=labels_format)


class VOCSemanticSegmentationDataProvider(VOCSegmentationDataProvider):
    def __init__(self, root: str, labels_format: Literal["yolo"] | None = None):
        super().__init__(root=root, mode="semantic", labels_format=labels_format)


class VOCPersonPartSegmentationDataProvider(VOCSegmentationDataProvider):
    URL = "http://liangchiehchen.com/projects/DeepLab.html"

    def __init__(self, root: str, labels_format: Literal["yolo"] | None = None):
        super().__init__(root=root, mode="person_part", labels_format=labels_format)

    def unzip(self, remove: bool = False):
        Path(self.raw_root).mkdir(parents=True, exist_ok=True)
        is_voc_present = False
        is_personpart_present = False

        for zip_filepath in self.zip_filepaths:
            if zip_filepath.endswith("trainval_2012.tar") and path_exists(
                f"{self.raw_root}/Annotations"
            ):
                log.info("VOC trainval 2012 already unzipped. Skipping")
                is_voc_present = True
                continue
            elif zip_filepath.endswith("person_part.zip") and path_exists(
                f"{self.raw_root}/ImageSets/SegmentationPersonPart"
            ):
                is_personpart_present = True
                log.info("PersonPart dataset already unzipped. Skipping")
                continue
            else:
                unzip(zip_filepath, self.root, remove)
        self.move_to_raw_root(is_voc_present, is_personpart_present)
        if remove:
            self.zip_filepaths.clear()

    def move_to_raw_root(self, is_voc_present: bool, is_personpart_present: bool):
        if not is_voc_present:
            super().move_to_raw_root()
        if is_personpart_present:
            return
        ids_src_dir = f"{self.root}/pascal_person_part/pascal_person_part_trainval_list"
        ids_dst_dir = f"{self.raw_root}/ImageSets/SegmentationPersonPart"
        move(ids_src_dir, ids_dst_dir)

        masks_src_dir = f"{self.root}/pascal_person_part/pascal_person_part_gt"
        masks_dst_dir = f"{self.raw_root}/SegmentationPersonPart"
        move(masks_src_dir, masks_dst_dir)
        remove_directory(f"{self.root}/pascal_person_part")


class VOCSemanticSegmentationAugDataProvider(VOCSegmentationDataProvider):
    def __init__(self, root: str, labels_format: Literal["yolo"] | None = None):
        super().__init__(root=root, mode="semantic_aug", labels_format=labels_format)

    def download(self):
        download_file(_TRAINAUG_IDS_URL, f"{self.root}/trainaug.txt")
        return super().download()

    def unzip(self, remove: bool = False):
        Path(self.raw_root).mkdir(parents=True, exist_ok=True)
        is_voc_present = False
        is_sbd_present = False
        if not path_exists(f"{self.root}/SegmentationClassAug.zip"):
            msg = (
                "No SegmentationClassAug.zip file in the root directory."
                + f"Download it from {_SBD_URL['trainaug']}, place it in the root and try again."
            )
            log.error(msg)
            raise FileNotFoundError(msg)

        self.zip_filepaths.append(f"{self.root}/SegmentationClassAug.zip")
        for zip_filepath in self.zip_filepaths:
            if zip_filepath.endswith("trainval_2012.tar") and path_exists(
                f"{self.raw_root}/Annotations"
            ):
                log.info("VOC trainval 2012 already unzipped. Skipping")
                is_voc_present = True
                continue
            elif zip_filepath.endswith("SegmentationClassAug.zip") and path_exists(
                f"{self.raw_root}/ImageSets/SegmentationClassAug"
            ):
                is_sbd_present = True
                log.info("SegmentationClassAug dataset already unzipped. Skipping")
                continue
            elif zip_filepath.endswith(("trainval_2012.tar", "SegmentationClassAug.zip")):
                unzip(zip_filepath, self.root, remove)
        self.move_to_raw_root(is_voc_present, is_sbd_present)
        if remove:
            self.zip_filepaths.clear()

    def move_to_raw_root(self, is_voc_present: bool, is_sbd_present: bool):
        if not is_voc_present:
            super().move_to_raw_root()
        if is_sbd_present:
            return
        masks_src_dir = f"{self.root}/SegmentationClassAug"
        masks_dst_dir = f"{self.raw_root}/SegmentationClassAug"
        move(masks_src_dir, masks_dst_dir)
        img_sets_dir = f"{self.raw_root}/ImageSets"
        Path(f"{img_sets_dir}/SegmentationClassAug").mkdir(parents=True, exist_ok=True)
        move(f"{self.root}/trainaug.txt", f"{img_sets_dir}/SegmentationClassAug/trainaug.txt")
        src_ids_files = [
            f"{img_sets_dir}/Segmentation/train.txt",
            f"{img_sets_dir}/Segmentation/val.txt",
        ]
        dst_ids_files = [
            f"{img_sets_dir}/SegmentationClassAug/train.txt",
            f"{img_sets_dir}/SegmentationClassAug/val.txt",
        ]
        copy_files(src_ids_files, dst_ids_files)
        # TODO: Add png colorization


if __name__ == "__main__":
    from geda.utils.config import ROOT

    root = str(ROOT / "data" / "voc_2012")
    # dp = VOCPersonPartSegmentationDataProvider(root)
    dp = VOCInstanceSegmentationDataProvider(root)
    # dp = VOCSemanticSegmentationDataProvider(root)



from PIL import Image
import numpy as np
import cv2
import glob
from tqdm.auto import tqdm


def find_polygons_in_binary_mask(binary_mask: np.ndarray) -> list[np.ndarray]:
    contours, _ = cv2.findContours(binary_mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
    polygons = [polygon.squeeze() for polygon in contours]
    return polygons


def parse_segmentation_mask_to_yolo_format(
    mask: np.ndarray,
    background: int | None = None,
    contour: int | None = None,
    return_str: bool = True,
) -> list[list[int | float]] | str:
    """Return list of YOLO like annotations for segmentation purposes, i.e.
    [
        [class_id, x1, y1, x2, y2, ..., xn, yn], # object_0
        [class_id, x1, y1, x2, y2, ..., xn, yn], # object_1
        ...,
        [class_id, x1, y1, x2, y2, ..., xn, yn] # object_m
    ]
    mask attribute is a
    """
    unique_label_ids = np.unique(mask).tolist()
    if background is not None and background in unique_label_ids:
        unique_label_ids.remove(background)  # remove background from labels
    if contour is not None and contour in unique_label_ids:
        unique_label_ids.remove(contour)  # remove contour from labels
    annotations = []
    for label_id in unique_label_ids:
        binary_mask = ((mask == label_id) * 1).astype(np.uint8)
        polygons = find_polygons_in_binary_mask(binary_mask)
        wh = np.flip(np.array(binary_mask.shape))  # for normalization purposes
        norm_polygons = [polygon / wh for polygon in polygons]
        xy_sequences = [polygon.flatten().tolist() for polygon in norm_polygons]
        for xy_sequence in xy_sequences:
            if len(xy_sequence) // 2 < 3:
                continue
            annotations.append([label_id] + xy_sequence)
    if return_str:
        annotations_str = [" ".join([str(el) for el in annot]) for annot in annotations]
        return "\n".join(annotations_str)
    return annotations


def parse_segmentation_masks_to_yolo(
    masks_filepaths: list[str],
    background: int | None = None,
    contour: int | None = None,
):
    for filename in tqdm(masks_filepaths, desc="Creating labels in yolo format"):
        mask = np.array(Image.open(filename))
        label_filepath = filename.replace("masks", "labels").replace(".png", ".txt")
        annot_txt = parse_segmentation_mask_to_yolo_format(
            mask=mask, background=background, contour=contour
        )
        with open(label_filepath, "w") as file:
            file.write(annot_txt)
